/*
 * Copyright (c) 2010 NTNU/KTH; author: Oleksandr Bodriagov. All rights reserved. :)
 
 * Class ApduProcessor: processes APDUs received from the PC
 */


package pkiphone;

public class ApduProcessor
{
    private CardCommunicator smartCard;

    final static String IDMP_AID = "A00000039742544659";
    final static String PIV_AID = "A00000030800001000";

    final static byte CLA_ISO7816 = (byte)0x00;
    final static byte CLA_MY = (byte)0x20;
    final static byte INS_SELECT = (byte)0xA4;
    final static byte INS_GET_DATA1 = (byte)0xCA;
    final static byte INS_GET_DATA2 = (byte)0xCB;
    final static byte INS_RESET = (byte)0x10;
    final static String SELECT_DF = "00A4000C023FFF";//select the parent DF (root of the application).

    final static byte [] SW_NO_ERROR = {(byte)0x90,(byte)0x00};
    final static byte [] SW_INS_NOT_SUPPORTED = {(byte)0x6D,(byte)0x00};
    final static byte [] SW_CLA_NOT_SUPPORTED = {(byte)0x6E,(byte)0x00};
    final static byte [] SW_NOT_FOUND = {(byte)0x6A,(byte)0x82};
    final static byte [] SW_INVALID = {(byte)0x6D,(byte)0x00}; //Instruction code not supported or invalid
    final static byte [] ANSWER_TO_RESET = {(byte)0x3b, (byte)0x89, (byte)0x40, (byte)0x14, (byte)0x47, (byte)0x47,
                                            (byte)0x32, (byte)0x34, (byte)0x4d, (byte)0x35, (byte)0x32, (byte)0x38, (byte)0x30};


    public ApduProcessor (CardCommunicator card)
    {
        smartCard = card;
    }


    public byte [] processApdu (byte[] receivedApdu, int dataLength)
    {
        if (receivedApdu[0] == CLA_ISO7816)
        {
            switch(receivedApdu[1])
            {
            case INS_SELECT:
                String hexApdu = L2CAPClient.getHex(receivedApdu, receivedApdu.length);
                if (hexApdu.regionMatches(true, 10, IDMP_AID, 0, IDMP_AID.length()))
                    return SW_NO_ERROR;
                else if (hexApdu.regionMatches(true, 10, PIV_AID, 0, PIV_AID.length()))
                    return SW_NOT_FOUND;
                else if (hexApdu.regionMatches(true, 0, SELECT_DF, 0, SELECT_DF.length()))
                    return SW_NOT_FOUND;
                else
                    return SW_NOT_FOUND;
            case INS_GET_DATA1:
                return SW_NOT_FOUND;
                //break;
            case INS_GET_DATA2:
                //return SW_NOT_FOUND;
                break;
            /*
            case INS_UPDATE_PIN:
                boolean updated = smartCard.updatePIN(truncatedApdu);
                return SW_NO_ERROR;
                              
            case INS_SIGN:
                byte [] signedData = smartCard.sign(truncatedApdu);
                return signedData;
             */
            default:
                return SW_INS_NOT_SUPPORTED;
            }
        }
        
        else if (receivedApdu[0] == CLA_MY)            
        {
            switch(receivedApdu[1])
            {
                case INS_RESET:
                    return ANSWER_TO_RESET;
                default:
                    return SW_INS_NOT_SUPPORTED;
            }
        }

        else return SW_CLA_NOT_SUPPORTED;

        /*
        String hexApdu = L2CAPClient.getHex(receivedApdu, 4);
        if (hexApdu.compareTo(apdu_SELECT) == 0 )
        */
        return null;

    }// end processApdu
}
