/*
 * Copyright (c) 2010 NTNU/KTH; author: Oleksandr Bodriagov. All rights reserved. :)
 
 * Class DiscoveryManager: responsible for performing a service search.
 *                         It receives a list of UUIDs to search for,
 *                         and replies with the found serviceRecords.
 */

package pkiphone;

import javax.bluetooth.*;
import java.util.*;

public class DiscoveryManager implements DiscoveryListener
{
     private Monitor boss;

     private UUID[] servicesList;//list of services that we are looking for


     private ServiceRecord[] serviceRecord;
     private Vector remoteBthDevices; // vector of discovered devices


     private LocalDevice localDevice;
     private DiscoveryAgent discoveryAgent;

     public DiscoveryManager (Monitor boss)
     {
        this.boss = boss;

        serviceRecord = null;
        servicesList = null;
        remoteBthDevices = new Vector();
     }

     //search routine
     public void performServiceSearch(UUID[] servList)
     {
         servicesList = servList;
         remoteBthDevices.removeAllElements();
         if (servList == null)
         {
             boss.errorInform("Error in performSearch: service list is empty");
             return;
         }

         try
         {
             localDevice = LocalDevice.getLocalDevice();
             discoveryAgent = localDevice.getDiscoveryAgent();

             discoveryAgent.startInquiry(DiscoveryAgent.GIAC, this);
         }
         catch (BluetoothStateException e)
         {
             boss.errorInform("Error in performSearch: " + e.toString());
         }
     }



     // device discovered callback
     public void deviceDiscovered(RemoteDevice btDevice, DeviceClass cod)
     {
         if (cod.getMajorDeviceClass() == 0x0100)
         remoteBthDevices.addElement(btDevice);// bluetooth device found
     }



     // device search completed callback
     public void inquiryCompleted(int discType)
     {
          if (!remoteBthDevices.isEmpty())
              searchServices(servicesList);
          else
              boss.errorInform("inquiryCompleted: No devices found");
     }

     

     // Searches for services on remote devices
     private void searchServices(UUID[] services)
     {
         for (int i = 0; i < remoteBthDevices.size(); i++)
         {
             try
             {
                 int transID;
                 transID = discoveryAgent.searchServices(null, services, (RemoteDevice)(remoteBthDevices.elementAt(i)), this);
             }
             catch (BluetoothStateException e)
             {
                 boss.errorInform("Error in searchServices: " + e.toString());
             }
         }
     }



     // service dicovered callback
     public void servicesDiscovered(int transID, ServiceRecord[] sRecord)
     {
         //if (serviceRecord != null)
           //  return;
         serviceRecord = sRecord;
     }



     // service search completed callback
     public void serviceSearchCompleted(int transID, int respCode)
     {
         boss.srvScanCompleted(serviceRecord);
     }






}
