
/*
 * Copyright (c) 2010 NTNU/KTH; author: Oleksandr Bodriagov. All rights reserved. :)

 *  Class L2CAPClient: handles GUI and all other classes
 */


package pkiphone;

import pkiphone.DiscoveryManager;
import pkiphone.CardCommunicator;
import pkiphone.ApduProcessor;
import javax.microedition.midlet.MIDlet;
import javax.microedition.lcdui.*;
import javax.microedition.io.Connector;
import javax.microedition.apdu.APDUConnection;
import java.util.*;

import javax.bluetooth.*;




public class L2CAPClient extends MIDlet implements CommandListener, ItemCommandListener, Runnable, Monitor //implements CommandListener, Runnable for starting new threads
 {
     private final String appletAID = "apdu:0;target=A0.00.00.00.62.91";

     //------------ GUI elements----------------------------------------------
     private Display mdisplay;
     private Form mForm;
     private Form waitForm;
     private Form errorForm;
     private Form exchangeForm;
     private Command connectCommand;
     private Command exitCommand;
     private Command backCommand;
     private Command cancelCommand;
     //private TextField pinField;
     private StringItem connect;
     private StringItem errLine;
     private TextField notificationText;
     private Spacer mSpace1;
     private Spacer mSpace2;
     private Gauge mGauge;


     //-------Bluetooth L2CAP part---------------------------------------------------------
     private DiscoveryManager serviceLocator;
     private ServiceRecord[] serviceRecord;
     private L2CAPConnection L2conn;
     private Receiver dataReceiver;

     private UUID[] servicesList;
     private final String BTHSCRD_UUID = "c07508f2b97043cab5ddcc4f2391bef4";
     private final String SERVICE_NAME = "BthScrdSrv";


     //-------Java Card part-------------------------------------------------------------------
     private ApduProcessor apduProcessor;
     private CardCommunicator cardCommunicator;
     private APDUConnection cardConnection;

     

     // Constructor
     public L2CAPClient()
     {
         //forms used by MIDlet
         mForm = new Form("L2CAP PKINIT");
         waitForm = new Form("L2CAP PKINIT");
         errorForm = new Form("L2CAP PKINIT");
         exchangeForm = new Form ("L2CAP PKINIT");


         exitCommand = new Command("Exit", Command.EXIT, 0);
         cancelCommand = new Command("Cancel", Command.CANCEL, 0);
         backCommand = new Command("Back", Command.BACK, 0);

         //main Form
         connect = new StringItem("", "Connect to PC",StringItem.BUTTON);
         mSpace1 = new Spacer (8,20);
         mSpace2 = new Spacer (8,10);
         connect.setLayout(Item.LAYOUT_CENTER | Item.LAYOUT_NEWLINE_BEFORE);
         connectCommand = new Command("",Command.ITEM,0);
         connect.setDefaultCommand(connectCommand);
         connect.setItemCommandListener(this);
         
         //pinField = new TextField("Enter PIN:", null, 8, TextField.PASSWORD | TextField.NUMERIC);
         //pinField.setLayout(Item.LAYOUT_VCENTER);
         mForm.append(mSpace1);
         //mForm.append(pinField);

         mForm.append(mSpace2);
         mForm.append(connect);
         mForm.addCommand(exitCommand);
         mForm.setCommandListener(this);

         //waitForm
         mGauge = new Gauge("Searching for Bluetooth devices\n\n", false, Gauge.INDEFINITE, Gauge.CONTINUOUS_RUNNING);
         mGauge.setLayout(Item.LAYOUT_CENTER | Item.LAYOUT_VCENTER);
         waitForm.addCommand(cancelCommand);
         waitForm.append(mGauge);
         waitForm.setCommandListener(this);

         //errorForm
         errLine = new StringItem("", "",StringItem.PLAIN);         
         errLine.setLayout(Item.LAYOUT_CENTER | Item.LAYOUT_NEWLINE_BEFORE);
         errorForm.append(errLine);
         errorForm.addCommand(backCommand);
         errorForm.setCommandListener(this);

         //APDU exchange form
         notificationText = new TextField("Exchange:", null, 800, TextField.UNEDITABLE);
         notificationText.setLayout(Item.LAYOUT_VCENTER);
         exchangeForm.append(notificationText);
         exchangeForm.addCommand(backCommand);
         exchangeForm.setCommandListener(this);



         //--------------------- L2CAP part --------------------------------
         serviceLocator = new DiscoveryManager(this);
         
         servicesList = new UUID[2];
         // Add the UUID for L2CAP to make sure that the service record found has L2CAP support
         servicesList[0] = new UUID(0x0100);// L2CAP UUID16: 0x0100
         servicesList[1] = new UUID(BTHSCRD_UUID, false);//Bluetooth Smart Card Reader UUID

         cardConnection = null;
     
     }// end constructor


     public void startApp()
     {
         mdisplay = Display.getDisplay(this);
         mdisplay.setCurrent(mForm);
     }

     public void pauseApp() {}
     public void destroyApp(boolean unconditional) {}


     //handling for standard commands
     public void commandAction(Command c, Displayable s)
     {
         if (c == exitCommand)
             notifyDestroyed();

         else if (c == backCommand)
         {
             if(dataReceiver != null)
                 dataReceiver.stop();
             
             //pinField.setString("");
             mdisplay.setCurrent(mForm);
         }

         else if (c == cancelCommand)
         {
             if(dataReceiver != null)
                 dataReceiver.stop();
             try {L2conn.close();}
             catch (Exception e)
             {
                  Form f = new Form("Exception:");
                  f.append(e.toString());
                  f.addCommand(backCommand);
                  f.setCommandListener(this);
                  mdisplay.setCurrent(f);
             }
             //pinField.setString("");
             mdisplay.setCurrent(mForm);
         }
     }


     //handling for commands associated with on-screen Items
     public void commandAction( Command cC, Item cItem )
     {
         if (cC == connectCommand && cItem == connect)
         {
            /* uncomment if everything is ready for communication with a real card
             try
             {
                 cardConnection = (APDUConnection)Connector.open(appletAID);
             }
             catch (Exception e)
             {
                 cardConnection = null;
                 errLine.setText("Error: Could not connect to the card");
                 mdisplay.setCurrent(errorForm);
                 return;
             }
             */

             cardCommunicator = new CardCommunicator(cardConnection);
                 
             Thread connectionThread = new Thread(this);//Bluetooth communication thread
             connectionThread.start();
         }
     }




     // (new thread): creates new APDU connection and exchanges data with the card
     public void run()
     {
         mGauge.setLabel("Searching for Bluetooth smart card reader\n");
         mdisplay.setCurrent(waitForm);

         serviceLocator.performServiceSearch(servicesList);
     }



     // service search completition callback
     public void srvScanCompleted(ServiceRecord[] sRecords)
     {
         serviceRecord = null;
         if (sRecords == null)
         {
             errLine.setText("No matching services found");
             mdisplay.setCurrent(errorForm);
             return;
         }
         serviceRecord = sRecords;

         DataElement serviceElem;
         String servName = null;
         for (int i = 0; i < serviceRecord.length; i++)
         {
             serviceElem = serviceRecord[i].getAttributeValue(0x0006);
             if (serviceElem != null)
             servName = (String)serviceElem.getValue();
         }
         if (servName != null)
             errLine.setText("Found service: " + servName );


         //
         // try to connect to the server
         String srvURL = serviceRecord[0].getConnectionURL(ServiceRecord.AUTHENTICATE_ENCRYPT, false);

         if (L2conn != null)
         {
             try
             {
                 L2conn.close();
             }
             catch (Exception e)
             {
                 Form f = new Form("Exception:");
                 f.append(e.toString());
                 f.addCommand(backCommand);
                 f.setCommandListener(this);
                 mdisplay.setCurrent(f);
             }
         }
         if (srvURL != null)
             startBthConnection(srvURL);
         else
         {
              errLine.setText("Service URL = NULL ");
              mdisplay.setCurrent(errorForm);
         }

     }// end srvScanCompleted



     private void startBthConnection(String serverURL )
     {
          if (serverURL == null)
          {
              errLine.setText("server URL is empty");
              mdisplay.setCurrent(errorForm);
              return;
          }

          errLine.setText(errLine.getText() + " Connecting to service...");
          mdisplay.setCurrent(errorForm);
          try
          {
              // connect to the server
              L2conn = (L2CAPConnection) Connector.open(serverURL);
              dataReceiver = new Receiver(this, L2conn);
              dataReceiver.start();
              apduProcessor = new ApduProcessor(cardCommunicator);
              errLine.setText("Connected !");
              mdisplay.setCurrent(errorForm);
          }
          catch (Exception ex)
          {
              errLine.setText("Found service. Connection  failed: " + ex.toString());
              mdisplay.setCurrent(errorForm);
          }

     }// end startBthConnection


     //
     public void errorInform(String errorDescription)
     {
         errLine.setText(errorDescription);
         mdisplay.setCurrent(errorForm);
     }



     public void receivedData (byte[] data, int length)
     {
         if(length > 0 && data != null)
         {
             byte[] response = apduProcessor.processApdu(data, length);
             String msg = getHex(data, length);
             String rcode = getHex(response, response.length);

             if (response != null)
                 try{
                     L2conn.send(response);
                 }
                 catch (Exception e)
                 {
                     errLine.setText("Exception in receivedData: " + e.toString());
                     mdisplay.setCurrent(errorForm);
                 }

             String exchange = "\n"+ msg + "; Res: " + rcode + "\n";
             int position = notificationText.size();
             notificationText.insert(exchange, position);
             mdisplay.setCurrent(exchangeForm);
         }
         else
         {
             errLine.setText("No data received");
             mdisplay.setCurrent(errorForm);
         }
     }



     public static String getHex( byte [] buf, int length )
     {
         if (buf == null || length == 0)
             return null;
         
         char[] HEXES = "0123456789ABCDEF".toCharArray();
         char[] chars = new char[2 * length];
         for (int i = 0; i < length; ++i)
         {
             chars[2 * i] = HEXES[(buf[i] & 0xF0) >>> 4];
             chars[2 * i + 1] = HEXES[buf[i] & 0x0F];
         }
         return new String(chars);
     }





}
