/*
 * Copyright (c) 2010 NTNU/KTH author: Oleksandr Bodriagov. All rights reserved. :)

 *  BrBServerContext.h

 *  contains server context structure and brb functions
 */

#ifndef BRBSRV_CONTEXT_H
#define BRBSRV_CONTEXT_H


#include <ntddk.h>
#include <wdf.h>
#include <initguid.h> 
#include <ntstrsafe.h>
#include <bthdef.h>
#include <ntintsafe.h>
#include <bthguid.h>
#include <bthioctl.h>
#include <sdpnode.h>
#include <bthddi.h>
#include <bthsdpddi.h>
#include <bthsdpdef.h>

#pragma warning(push)
#pragma warning(disable:4201)
#include "smclib.h"
#pragma warning(pop)

#define POOLTAG_BTHECHOSAMPLE 'htbw'


typedef struct _BTH_DEVICE_CONTEXT
{
	SMARTCARD_EXTENSION SmartCardExtension;
	
	//----------------------------
	WDFDEVICE Device;

	// Default I/O target
    WDFIOTARGET IoTarget;

    // Profile driver interface which contains profile driver DDI
    BTH_PROFILE_DRIVER_INTERFACE ProfileDrvInterface;

    // Local Bluetooth Address
    BTH_ADDR LocalBthAddr;

    // Preallocated request to be reused during initialization/deinitialzation phase
    // Access to this reqeust is not synchronized
    WDFREQUEST Request;

	//------------------------

    // The Data Protocol/Service Multiplexer (PSM) that the channel uses to connect to the remote device  
    USHORT Psm;

    // Handle to published SDP record
    HANDLE_SDP SdpRecordHandle;

    // Handle obtained by registering L2CAP server
    L2CAP_SERVER_HANDLE L2CAPServerHandle;

    /*
    BRB used for server and PSM register and unregister

	Server and PSM register and unregister must be done sequentially
	since access to this brb is not synchronized.*/

    struct _BRB RegisterUnregisterBrb;
    
    // Connection List lock
    WDFSPINLOCK ConnectionListLock;

    // Outstanding open connections
    LIST_ENTRY ConnectionList;

	WDFQUEUE IoctlQueue;
	WDFQUEUE NotificationQueue;
	ULONG DeviceInstanceNo;

	KEVENT ReceivedResponseEvt;//synchronization event: waiting for bluetooth client response
	BOOLEAN WaitingResponse;

    
} BTH_DEVICE_CONTEXT, *PBTH_DEVICE_CONTEXT;




__drv_maxIRQL(PASSIVE_LEVEL)
NTSTATUS
//FORCEINLINE
BthEchoSharedSendBrbSynchronously(
    __in WDFIOTARGET IoTarget,
    __in WDFREQUEST Request,
    __in PBRB Brb,
    __in ULONG BrbSize
    );



__drv_maxIRQL(DISPATCH_LEVEL)
NTSTATUS
BthEchoSharedSendBrbAsync(
    __in WDFIOTARGET IoTarget,
    __in WDFREQUEST Request,
    __in PBRB Brb,
    __in size_t BrbSize,
    __in PFN_WDF_REQUEST_COMPLETION_ROUTINE ComplRoutine,
    __in_opt WDFCONTEXT Context
    );


#endif