/*++

Copyright (c) 1997 - 1999 SCM Microsystems, Inc.

Module Name:
    bthpcsc.h

Abstract:
    Driver header - NT Version

Author:

    Andreas Straub  (SCM Microsystems, Inc.)
    Klaus Schuetz   (Microsoft Corp.)

--*/

#if !defined ( __PSCR_NT_DRV_H__ )
#define __PSCR_NT_DRV_H__
#define SMARTCARD_POOL_TAG '4SCS'



#include "brbservercontext.h"
#include "connection.h"
#include "guid.h"
#include "pcscrw.h"

#include <DEVIOCTL.H>


__kernel_driver

#if !defined( STATUS_DEVICE_REMOVED )
#define STATUS_DEVICE_REMOVED STATUS_UNSUCCESSFUL
#endif

#define SysCompareMemory( p1, p2, Len )         ( RtlCompareMemory( p1,p2, Len ) != Len )
#define SysCopyMemory( pDest, pSrc, Len )       RtlCopyMemory( pDest, pSrc, Len )
#define SysFillMemory( pDest, Value, Len )      RtlFillMemory( pDest, Len, Value )

#define DELAY_WRITE_PSCR_REG    1
#define DELAY_PSCR_WAIT         5

#define LOBYTE( any )   ((UCHAR)( any & 0xFF ) )
#define HIBYTE( any )   ((UCHAR)( ( any >> 8) & 0xFF ))

//
// These macros will store and retrieve the request handle from the IRP.
// We will use the first slot of DriverContext because framework reserves 
// the use of 3rd & 4rd slot of DriverContext for its own purpose.
//
#define GET_WDFREQUEST_FROM_IRP(Irp) (WDFREQUEST)Irp->Tail.Overlay.DriverContext[0]
#define SET_REQUEST_IN_IRP(Irp, Request) {Irp->Tail.Overlay.DriverContext[0] = (PVOID)Request;}


WDF_DECLARE_CONTEXT_TYPE_WITH_NAME(BTH_DEVICE_CONTEXT, GetServerDeviceContext)

#define IOCTL_PSCR_COMMAND      SCARD_CTL_CODE( 0x8000 )
#define IOCTL_GET_VERSIONS      SCARD_CTL_CODE( 0x8001 )
#define IOCTL_SET_TIMEOUT       SCARD_CTL_CODE( 0x8002 )
#define IOCTL_GET_CONFIGURATION SCARD_CTL_CODE( 0x8003 )


typedef struct _VERSION_CONTROL
{
    ULONG   SmclibVersion;
    UCHAR   DriverMajor,
            DriverMinor,
            FirmwareMajor,
            FirmwareMinor,
            UpdateKey;
} VERSION_CONTROL, *PVERSION_CONTROL;

#define SIZEOF_VERSION_CONTROL  sizeof( VERSION_CONTROL )


typedef struct _PSCR_CONFIGURATION
{
    PPSCR_REGISTERS IOBase;
    ULONG           IRQ;

} PSCR_CONFIGURATION, *PPSCR_CONFIGURATION;

#define SIZEOF_PSCR_CONFIGURATION   sizeof( PSCR_CONFIGURATION )



DRIVER_INITIALIZE DriverEntry;

EVT_WDF_DRIVER_DEVICE_ADD PcscEvtDeviceAdd;

EVT_WDF_OBJECT_CONTEXT_CLEANUP PcscEvtDeviceContextCleanup;


EVT_WDF_DEVICE_SELF_MANAGED_IO_INIT BthEchoSrvEvtDeviceSelfManagedIoInit;

EVT_WDF_DEVICE_SELF_MANAGED_IO_CLEANUP BthEchoSrvEvtDeviceSelfManagedIoCleanup;


EVT_WDF_IO_QUEUE_IO_DEVICE_CONTROL PscrEvtIoDeviceControl;
EVT_WDF_IO_QUEUE_IO_CANCELED_ON_QUEUE PscrEvtIoCanceledOnQueue;

NTSTATUS
BthScrdEvtDeviceSelfManagedIoInit(
    __in WDFDEVICE  Device
    );

VOID
BthScrdEvtDeviceSelfManagedIoCleanup(
    __in WDFDEVICE  Device
    );

NTSTATUS
PcscRegisterWithSmcLib(
    PBTH_DEVICE_CONTEXT pDeviceContext
    );


NTSTATUS
complete_SmartCardIoctl(
	 IN WDFQUEUE Queue,
	 IN WDFREQUEST Request,
	 IN size_t OutputBufferLength,
	 IN size_t InputBufferLength,
	 IN ULONG IoControlCode);




__drv_functionClass(IO_COMPLETION_ROUTINE)
NTSTATUS
PscrSmcLibComplete (
    IN PDEVICE_OBJECT DeviceObject,
    IN PIRP Irp,
    IN PVOID Context
    );



PCHAR
GetIoctlName(
    ULONG IoControlCode
    );

#endif  // __PSCR_NT_DRV_H__

