/*++

Copyright (c) 1997 - 1999 SCM Microsystems, Inc.

Module Name:
    PscrRdWr.h

Abstract:
    Constants & access function prototypes for SCM PSCR smartcard reader

Author:
    Andreas Straub

Revision History:
    Oleksandr Bodriagov          2010   changed

--*/

#if !defined( __PSCR_RDWR_H__ )
#define __PSCR_RDWR_H__

#include <ntddk.h>
#include <wdf.h>

#pragma pack( 1 )
//
//  The usage of the PSCR_REGISTERS struct is a little bit tricky:
//  We set the address of that stucture to the IO Base Port, then
//  the other reg's can accessed by their address.
//  p.E.    &PscrRegs = 0x320 --> &PscrRegs->CmdStatusReg = 0x321...
//

// evil hack :)
//typedef unsigned char UCHAR, *PUCHAR;
//typedef unsigned char ULONG, *PULONG;
typedef struct _PSCR_REGISTERS {

    UCHAR  DataReg;
    UCHAR  CmdStatusReg;
    UCHAR  SizeLSReg;
    UCHAR  SizeMSReg;

 } PSCR_REGISTERS, *PPSCR_REGISTERS;

#define SIZEOF_PSCR_REGISTERS       ( sizeof( PSCR_REGISTERS  )

#pragma pack()


typedef enum _READER_POWER_STATE {
    PowerReaderUnspecified = 0,
    PowerReaderWorking,
    PowerReaderOff
} READER_POWER_STATE, *PREADER_POWER_STATE;

typedef void ( *PTRACKING_COMPLETION )( PVOID TrackingContext, UCHAR Status, BOOLEAN Wakeup );

#define MAX_DATARATES   4



typedef struct _READER_EXTENSION {

    //  I/O address where the reader is configured.
    PPSCR_REGISTERS IOBase;
    ULONG IOWindow;
    ULONG MaxIFSD;

    //  IRQ assigned by the system
    ULONG CurrentIRQ;

    //
    //  limit for read timeout. the absolute timeout limit is
    //  MaxRetries * DELAY_PSCR_WAIT
    //

    ULONG MaxRetries;

    //  Source/Destination byte always used by the PCMCIA (PC Card) reader.
    UCHAR Device;

    //  Software revision ID of the SwapSmart firmware.
    UCHAR FirmwareMajor, FirmwareMinor, UpdateKey;

    //
    //  Interrupt status; the flag is raised if an freeze event was detected
    //  and cleared if the interface was cleared (int service routine or
    //  PscrRead)
    //
    BOOLEAN FreezePending;

    //
    //  Selected file in the reader file system; the flag is raised if the
    //  ICC1 status file was selected & is cleared if any generic ioctl was
    //  issued (in this case another file may be selected)
    //
    BOOLEAN StatusFileSelected;

    //  used to cancel a read request
    BOOLEAN         RequestPending;
    //    set TRUE if an interrupt is sending, but the reader is busy
    BOOLEAN         RequestInterrupt;

    BOOLEAN InvalidStatus;

    BOOLEAN CardPresent;

    // Current reader power state.
    READER_POWER_STATE ReaderPowerState;

    PTRACKING_COMPLETION    CompletionRoutine;

    PVOID                   TrackingContext;

    ULONG                   dataRatesSupported[MAX_DATARATES];

} READER_EXTENSION, *PREADER_EXTENSION;

#define SIZEOF_READER_EXTENSION     ( sizeof( READER_EXTENSION ))


//
//  Constants -----------------------------------------------------------------
//

#define TLV_BUFFER_SIZE             0x20

#define CLA_RESET                   0x20
#define CLA_WARM_RESET              0x20
#define INS_RESET                   0x10
#define INS_WARM_RESET              0x1F



#endif  //  __PSCR_RDWR_H__



