/*
 * Copyright (c) 2010 NTNU/KTH author: Oleksandr Bodriagov. All rights reserved. :)

 *  ScardExtCallback.c

 *  contains callback functions for the smart card library SMCLIB
 */


#include "SCardExtCallback.h"
#include "serverblabla.h"





NTSTATUS
CBTransmit(
          PSMARTCARD_EXTENSION SmartcardExtension
          )
/*++
CBTransmit: SMCLIB RDF_TRANSMIT callback function

Return Value:
        STATUS_SUCCESS
        STATUS_NO_MEDIA
        STATUS_TIMEOUT
        STATUS_INVALID_DEVICE_REQUEST

--*/
{
    NTSTATUS  NTStatus = STATUS_SUCCESS;

	KdPrint(("  SCRD! CBTransmit: Enter\n"));

    //      dispatch on the selected protocol
    switch ( SmartcardExtension->CardCapabilities.Protocol.Selected )
	{
		case SCARD_PROTOCOL_T0:
			NTStatus = CBT0Transmit( SmartcardExtension );
			break;
		default:
			NTStatus = STATUS_INVALID_DEVICE_REQUEST;
			break;
    }
	KdPrint(("  SCRD! CBTransmit: Exit\n"));

    return( NTStatus );
}



NTSTATUS
CBCardPower(
           PSMARTCARD_EXTENSION SmartcardExtension
           )
/*++
CBCardPower:  callback handler for SMCLIB RDF_CARD_POWER

Return Value:
        STATUS_SUCCESS
        STATUS_NO_MEDIA
        STATUS_TIMEOUT
        STATUS_BUFFER_TOO_SMALL
--*/
{
	NTSTATUS     NTStatus = STATUS_SUCCESS;
    UCHAR        CommandBuffer[5];
    ULONG        i, Command, ATRLength = 0;
    BYTE         CardState;
    KIRQL        irql;

	PLIST_ENTRY ple;
	PBTH_DEVICE_CONTEXT pDevContext;
	PBTHECHO_CONNECTION connection;
	LARGE_INTEGER ResponseTimeout;//the time that the driver waits for the mobile phone to answer


#if DBG || DEBUG
    static PCHAR request[] = { "PowerDown",  "ColdReset", "WarmReset"};
#endif

	ResponseTimeout.QuadPart = -5*1000000;// == 500 ms
	pDevContext = GetServerDeviceContext(WdfWdmDeviceGetWdfDeviceHandle(SmartcardExtension->OsData->DeviceObject));
	ple = pDevContext->ConnectionList.Flink;
	connection = CONTAINING_RECORD(ple, BTHECHO_CONNECTION, ConnectionListEntry);

	// build reset command
	CommandBuffer [0] = 0x20; //CLA = 
	CommandBuffer [1] = 0x10; //INS = 
	CommandBuffer [2] = 0x01; //P1 =
	CommandBuffer [3] = 0x01; //P2 =
	CommandBuffer [4] = 0x00; //Lc =

	KdPrint(("  SCRD! CBCardPower: Enter, Request = %s\n", request[SmartcardExtension->MinorIoControlCode]));

    Command = SmartcardExtension->MinorIoControlCode;


    switch ( Command ) 
	{
    case SCARD_WARM_RESET:

        // if the card was not powerd, fall thru to cold reset
        KeAcquireSpinLock(&SmartcardExtension->OsData->SpinLock, &irql);

        if ( SmartcardExtension->ReaderCapabilities.CurrentState > SCARD_SWALLOWED ) 
		{
            KeReleaseSpinLock(&SmartcardExtension->OsData->SpinLock, irql);

            // reset the card

			for (i = 0; i< 5; i++)
				KdPrint(("  SCRD! CBCardPower: WARM_RESET [%d] = 0x%x\n", i, CommandBuffer[i] ));  

			// sends data in the buffer
			BthEchoSrvSendEcho(
				pDevContext->Device,
				pDevContext->IoTarget,
				connection,
				&CommandBuffer,
				5 //command length
				);

			KeClearEvent( &pDevContext->ReceivedResponseEvt);// set event to non-signaled state
			pDevContext->WaitingResponse = TRUE;
			
			NTStatus = KeWaitForSingleObject(
				&pDevContext->ReceivedResponseEvt,
				Executive,
				KernelMode,
				FALSE,
				&ResponseTimeout
				);
			pDevContext->WaitingResponse = FALSE;//either the response is sastisfied or it timed-out

			if ( NT_SUCCESS( NTStatus ))
			{
				KdPrint(("  SCRD! CBCardPower: KeWaitForSingleObject: success\n"));
				ATRLength = SmartcardExtension->SmartcardReply.BufferLength;
			}

            break;
		} 
		else //goes to COLD RESET
            KeReleaseSpinLock(&SmartcardExtension->OsData->SpinLock, irql);


    case SCARD_COLD_RESET:

        // reset the card

		for (i = 0; i< 5; i++)
			KdPrint(("  SCRD! CBCardPower: COLD_RESET [%d] = 0x%x\n", i, CommandBuffer[i] ));  

		// sends data in the buffer
		BthEchoSrvSendEcho(
			pDevContext->Device,
			pDevContext->IoTarget,
			connection,
			&CommandBuffer,
			5 //command length
			);

		KeClearEvent( &pDevContext->ReceivedResponseEvt);// set event to non-signaled state
		pDevContext->WaitingResponse = TRUE;
		
		NTStatus = KeWaitForSingleObject(
			&pDevContext->ReceivedResponseEvt,
			Executive,
			KernelMode,
			FALSE,
			&ResponseTimeout
			);
		pDevContext->WaitingResponse = FALSE;//either the response is sastisfied or it timed-out

		if ( NT_SUCCESS( NTStatus ))
		{
			KdPrint(("  SCRD! CBCardPower: KeWaitForSingleObject: success\n"));
			ATRLength = SmartcardExtension->SmartcardReply.BufferLength;
		}

        break;

    case SCARD_POWER_DOWN:
		ATRLength = 0;
        
		// some power down command to transmit to the mobile phone 
		NTStatus = STATUS_SUCCESS;

        // change card status
        CBUpdateCardState(SmartcardExtension, PSCR_ICC_ABSENT, FALSE);
        break;
    }

	


    if ( NT_SUCCESS( NTStatus ))  //received response
	{
		// update all neccessary data if an ATR was received
        if ( ATRLength > 2 ) 
		{
            //  copy ATR to user buffer buffer
            if ( ATRLength <= SmartcardExtension->IoRequest.ReplyBufferLength ) 
			{
                SysCopyMemory(
                             SmartcardExtension->IoRequest.ReplyBuffer,
                             SmartcardExtension->SmartcardReply.Buffer,
                             ATRLength
                             );
                *SmartcardExtension->IoRequest.Information = ATRLength;
            }

            //  copy ATR to card capability buffer
            if ( ATRLength <= MAXIMUM_ATR_LENGTH )
			{
                SysCopyMemory(
                             SmartcardExtension->CardCapabilities.ATR.Buffer,
                             SmartcardExtension->SmartcardReply.Buffer,
                             ATRLength
                             );

                SmartcardExtension->CardCapabilities.ATR.Length = ( UCHAR )ATRLength;

                 // let the lib update the card capabilities
                NTStatus = SmartcardUpdateCardCapabilities( SmartcardExtension );
            }
			else 
                NTStatus = STATUS_BUFFER_TOO_SMALL;

        }//end if ( ATRLength > 2 )
    }

    if ( !NT_SUCCESS( NTStatus )) 
	{
		switch ( NTStatus )
		{
			case STATUS_NO_MEDIA:
			case STATUS_BUFFER_TOO_SMALL:
				break;

			case STATUS_TIMEOUT:
				NTStatus = STATUS_IO_TIMEOUT;
				break;

			default:
				NTStatus = STATUS_UNRECOGNIZED_MEDIA;
				break;
		}
		if( NTStatus != STATUS_SUCCESS )
		{

		   // resman requires status ok to translate ntstatus to scard... error.
			if( SmartcardExtension->ReaderCapabilities.CurrentState <= SCARD_ABSENT )
			{
				SmartcardUpdateCardCapabilities( SmartcardExtension );
				NTStatus = STATUS_SUCCESS;
			}
		}
	}// end if ( !NT_SUCCESS( NTStatus )) 

	RtlZeroMemory (CommandBuffer, 5);// Clearing the buffers

	KdPrint(("  SCRD! CBCardPower: Exit, status (%lx)\n", NTStatus));
    return( NTStatus );
}





NTSTATUS
CBSetProtocol(
             PSMARTCARD_EXTENSION SmartcardExtension
             )
{

    //static PCHAR request[] = { "SCARD_PROTOCOL_UNDEFINED",  "SCARD_PROTOCOL_T0", "SCARD_PROTOCOL_T1", "SCARD_PROTOCOL_Tx"};


	NTSTATUS NTStatus = STATUS_PENDING;
    USHORT SCLibProtocol;
    UCHAR TLVList[ TLV_BUFFER_SIZE ];
    PREADER_EXTENSION ReaderExtension = SmartcardExtension->ReaderExtension;
    KIRQL irql;

    KeAcquireSpinLock(&SmartcardExtension->OsData->SpinLock, &irql);

    if (SmartcardExtension->ReaderCapabilities.CurrentState == SCARD_SPECIFIC) 
	{
        KeReleaseSpinLock(&SmartcardExtension->OsData->SpinLock, irql);

        return STATUS_SUCCESS;
    }
	else
		KeReleaseSpinLock(&SmartcardExtension->OsData->SpinLock, irql);


	KdPrint(("  SCRD! CBSetProtocol: Enter, Protocol = %s\n", SmartcardExtension->MinorIoControlCode));

    SCLibProtocol = ( USHORT )( SmartcardExtension->MinorIoControlCode );

	if (SCLibProtocol & (SCARD_PROTOCOL_T0 | SCARD_PROTOCOL_T1))
		NTStatus = STATUS_SUCCESS;

	else                 
        NTStatus = STATUS_INVALID_DEVICE_REQUEST;//      we don't support other modi

     // if protocol selection failed, prevent from calling invalid protocols
    if ( NT_SUCCESS( NTStatus ))
	{
        KeAcquireSpinLock(&SmartcardExtension->OsData->SpinLock, &irql);
        SmartcardExtension->ReaderCapabilities.CurrentState = SCARD_SPECIFIC;

		SCLibProtocol = (SCLibProtocol & SCARD_PROTOCOL_T1 &
                         SmartcardExtension->CardCapabilities.Protocol.Supported) ?
                        SCARD_PROTOCOL_T1 : SCARD_PROTOCOL_T0;

        KeReleaseSpinLock(&SmartcardExtension->OsData->SpinLock,  irql);
    } 
	else
        SCLibProtocol = SCARD_PROTOCOL_UNDEFINED;

    //Return the selected protocol to the caller.
    SmartcardExtension->CardCapabilities.Protocol.Selected = SCLibProtocol;
    *( PULONG )( SmartcardExtension->IoRequest.ReplyBuffer ) = SCLibProtocol;
    *( SmartcardExtension->IoRequest.Information ) = sizeof( ULONG );

	KdPrint(("  SCRD! CBSetProtocol: Exit, status = (%lx)\n", NTStatus));

    return( NTStatus );

}




NTSTATUS
CBCardTracking(
              PSMARTCARD_EXTENSION SmartcardExtension
              )
/*++
CBCardTracking:
        callback handler for SMCLIB RDF_CARD_TRACKING. the requested event was
        validated by the smclib (i.e. a card removal request will only be passed
        if a card is present).
        for NT the cancel routine for the irp will be set to the drivers cancel
        routine.

Arguments:
        SmartcardExtension      context of call

Return Value:
        STATUS_PENDING
--*/
{
    WDFREQUEST request;

	KIRQL CancelIrql;
	PBTH_DEVICE_CONTEXT pDevContext;
	NTSTATUS status = STATUS_SUCCESS;

	KdPrint(("  PSCR! CBCardTracking: Enter\n"));
	KdPrint(("  ---------------------------\n"));
	
	pDevContext = GetServerDeviceContext(WdfWdmDeviceGetWdfDeviceHandle(SmartcardExtension->OsData->DeviceObject));


	if(NT_SUCCESS(status))
	{
		IoAcquireCancelSpinLock(&CancelIrql);

		//sets up a driver-supplied Cancel routine to be called if a given IRP is canceled
		IoSetCancelRoutine(SmartcardExtension->OsData->NotificationIrp, VdCancel);
		IoReleaseCancelSpinLock(CancelIrql);
	}

    request = GET_WDFREQUEST_FROM_IRP(SmartcardExtension->OsData->NotificationIrp);

    pDevContext = GetServerDeviceContext(WdfIoQueueGetDevice(WdfRequestGetIoQueue(request)));

    IoMarkIrpPending(SmartcardExtension->OsData->NotificationIrp);

    //
    // Move the stack pointer back to where it was when the IRP was delivered to the driver.
	// This compensates for the IoSetNextIrpStackLocation that we did when we presented the IRP 
    // to the smartcard libarary. 
    //
    IoSkipCurrentIrpStackLocation(SmartcardExtension->OsData->NotificationIrp);

    status = WdfRequestForwardToIoQueue(request,
                               pDevContext->NotificationQueue);
    if (!NT_SUCCESS(status))
	{
		KdPrint(("PSCR!CBCardTracking: WdfRequestForwardToIoQueue failed, Status code %!STATUS!\n", status));
		ASSERT(NT_SUCCESS(status));
		//sets the address pointed to by the SmartcardExtension->OsData->NotificationIrp
		//parameter to the address that is the value of the NULL parameter
        InterlockedExchangePointer(
                                 &(SmartcardExtension->OsData->NotificationIrp),
                                 NULL
                                 );
        WdfRequestComplete(request, status);//To complete an I/O request
    }

    status = STATUS_PENDING;

	KdPrint(("  PSCR! CBCardTracking: Exit \n" ));

    return status;
}




NTSTATUS
VdCancel(
		 IN PDEVICE_OBJECT DeviceObject,
		 IN PIRP Irp
		 )
{
	PBTH_DEVICE_CONTEXT pDevContext;
	PSMARTCARD_EXTENSION smartcardExtension;
	pDevContext = GetServerDeviceContext(DeviceObject);
	smartcardExtension = &pDevContext->SmartCardExtension;

	KdPrint(("PSCR! VdCancel: Enter \n" ));

	ASSERT(Irp == smartcardExtension->OsData->NotificationIrp);

	smartcardExtension->OsData->NotificationIrp = NULL;

	Irp->IoStatus.Information = 0;
	Irp->IoStatus.Status = STATUS_CANCELLED;

	IoReleaseCancelSpinLock(Irp->CancelIrql);

	IoCompleteRequest(Irp,IO_NO_INCREMENT);

	return STATUS_CANCELLED;
}



NTSTATUS
CBT0Transmit(
            PSMARTCARD_EXTENSION SmartcardExtension
            )
/*++
CBT0Transmit:
        finishes the callback RDF_TRANSMIT for the T0 protocol

Arguments:
        SmartcardExtension      context of call

Return Value:
        STATUS_SUCCESS
        STATUS_NO_MEDIA
        STATUS_TIMEOUT
        STATUS_INVALID_DEVICE_REQUEST
--*/
{
    NTSTATUS NTStatus = STATUS_SUCCESS;
    PUCHAR pRequest;
	ULONG i;

	PLIST_ENTRY ple;
	PBTH_DEVICE_CONTEXT pDevContext;
	PBTHECHO_CONNECTION connection;
	LARGE_INTEGER ResponseTimeout;
	ResponseTimeout.QuadPart = -5*1000000;// == 500 ms
	pDevContext = GetServerDeviceContext(WdfWdmDeviceGetWdfDeviceHandle(SmartcardExtension->OsData->DeviceObject));
	ple = pDevContext->ConnectionList.Flink;
	connection = CONTAINING_RECORD(ple, BTHECHO_CONNECTION, ConnectionListEntry);

	KdPrint(("  SCRD! CBT0Transmit: Enter \n" ));

    pRequest = SmartcardExtension->SmartcardRequest.Buffer;
	

    // SmartcardT0Request: copy data (APDU)from the user buffer to the Buffer member
	// of the SmartcardRequest member of the SMARTCARD_EXTENSION structure pointed to by
	// SmartcardExtension; adjust the value in BufferLength member of the SmartcardRequest
	// member. The driver must send this buffer to the smart card.
    NTStatus = SmartcardT0Request( SmartcardExtension );

	if (NTStatus != STATUS_SUCCESS)
	{
		KdPrint(("  SCRD! CBT0Transmit: SmartcardT0Request, status = (%lx)\n", NTStatus ));
        return NTStatus;
	}

	for (i = 0; i< SmartcardExtension->SmartcardRequest.BufferLength; i++)
		KdPrint(("  SCRD! CBT0Transmit: Request [%d] = 0x%x\n", i, pRequest[i] ));  

	// sends data in the buffer
	BthEchoSrvSendEcho(
		pDevContext->Device,
		pDevContext->IoTarget,
		connection,
		SmartcardExtension->SmartcardRequest.Buffer,
		SmartcardExtension->SmartcardRequest.BufferLength
		);

	KeClearEvent( &pDevContext->ReceivedResponseEvt);// set event to non-signaled state
	pDevContext->WaitingResponse = TRUE;
	
	NTStatus = KeWaitForSingleObject(
		&pDevContext->ReceivedResponseEvt,
		Executive,
		KernelMode,
		FALSE,
		&ResponseTimeout
		);

	if ( NTStatus == STATUS_SUCCESS)//received response
	{
		// copies data from the receive buffer to the user buffer and checks the T=0 protocol status
		NTStatus = SmartcardT0Reply( SmartcardExtension );
		KdPrint(("  SCRD! CBT0Transmit: KeWaitForSingleObject: success\n"));
	}
	// status mapping between KeWaitForSingleObject and CBT0Transmit (RDF_TRANSMIT callback)
	else if (NTStatus == STATUS_TIMEOUT)
		NTStatus = STATUS_IO_TIMEOUT;
	else
		NTStatus = STATUS_NO_MEDIA;


	pDevContext->WaitingResponse = FALSE;//either the response is sastisfied or it timed-out

	KdPrint(("  SCRD! CBT0Transmit: Exit (%lx)\n", NTStatus ));

    return( NTStatus );
}



VOID
CBUpdateCardState(
                 PSMARTCARD_EXTENSION SmartcardExtension,
                 UCHAR IccState,
                 BOOLEAN SystemWakeUp
                 )
{
    ULONG oldState;
    NTSTATUS status;
    KIRQL irql;
    WDFREQUEST request;
    PBTH_DEVICE_CONTEXT pDevContext;


    KeAcquireSpinLock( &SmartcardExtension->OsData->SpinLock, &irql);

	KdPrint(("SCRD! CBUpdateCardState: Enter\n"));

    oldState = (SmartcardExtension->ReaderCapabilities.CurrentState > SCARD_ABSENT ?
     SCARD_PRESENT : SCARD_ABSENT);

    SmartcardExtension->ReaderCapabilities.CurrentState =
    (IccState == PSCR_ICC_PRESENT ? SCARD_PRESENT : SCARD_ABSENT);

	KdPrint(("SCRD! CBUpdateCardState: Smart card is %s\n",
                    IccState == PSCR_ICC_PRESENT ? "inserted" : "removed"));

    if ( SmartcardExtension->OsData->NotificationIrp != NULL &&
         (SystemWakeUp &&
          (oldState == SCARD_PRESENT || SmartcardExtension->ReaderCapabilities.CurrentState == SCARD_PRESENT) ||
          SmartcardExtension->ReaderCapabilities.CurrentState != oldState)) 
	{
        PIRP notificationIrp ;

		KdPrint(("SCRD! CBUpdateCardState: notification IRP = %p\n", SmartcardExtension->OsData->NotificationIrp));
		KdPrint(("SCRD! CBUpdateCardState: old state = %p\n", oldState));
		KdPrint(("SCRD! CBUpdateCardState: current state = %p\n", SmartcardExtension->ReaderCapabilities.CurrentState));

        notificationIrp = InterlockedExchangePointer(
                                 &(SmartcardExtension->OsData->NotificationIrp),
                                 NULL
                                 );

		pDevContext = GetServerDeviceContext(WdfWdmDeviceGetWdfDeviceHandle(SmartcardExtension->OsData->DeviceObject));

        KeReleaseSpinLock(
             &SmartcardExtension->OsData->SpinLock,
             irql
             );

		KdPrint(("SCRD! CBUpdateCardState: WdfIoQueueRetrieveNextRequest\n"));
		//retrieves the next available I/O request from a specified I/O queue.
        status = WdfIoQueueRetrieveNextRequest(
                            pDevContext->NotificationQueue,
                            &request);
        if (NT_SUCCESS(status))
		{
			KdPrint(("SCRD! CBUpdateCardState: Completing Irp %p\n", notificationIrp));


			//stores completion information and then completes a specified I/O request
			//with a supplied completion status
            WdfRequestCompleteWithInformation(request, status, 0);
        }
        else 
		{
            ASSERTMSG("!!!! CBUpdateCardState: WdfIoQueueRetrieveNextRequest failed", status == STATUS_NO_MORE_ENTRIES);
        }
    } 
	else 
	{
		KdPrint(("SCRD! CBUpdateCardState: notification IRP = %p\n", SmartcardExtension->OsData->NotificationIrp));
		KdPrint(("SCRD! CBUpdateCardState: old state = %p\n", oldState));
		KdPrint(("SCRD! CBUpdateCardState: current state = %p\n", SmartcardExtension->ReaderCapabilities.CurrentState));

        KeReleaseSpinLock(&SmartcardExtension->OsData->SpinLock, irql);
    }

	KdPrint(("SCRD! CBUpdateCardState: Exit\n"));
}
