/*
 * Copyright (c) 2010 NTNU/KTH author: Oleksandr Bodriagov. All rights reserved. :)

 *  Serverblabla.c

 *  contains function for Bluetooth server part
 */

#include "serverblabla.h"
#include "brbservercontext.h"
#include "bthpcsc.h"
#include "SCardExtCallback.h"


#ifdef ALLOC_PRAGMA
#pragma alloc_text (PAGE, BthSrvUnregisterPSM)
#pragma alloc_text (PAGE, BthSrvUnregisterL2CAPServer)
#pragma alloc_text (PAGE, BthEchoSrvPublishSdpRecord)
#pragma alloc_text (PAGE, BthSrvRemoveSdpRecord)
#endif




__drv_maxIRQL(DISPATCH_LEVEL)
void
BthEchoSrvIndicationCallback(
    __in PVOID Context,
    __in INDICATION_CODE Indication,
    __in PINDICATION_PARAMETERS Parameters
    )

/*++
Description:
    Indication callback passed to bth stack while registering server.
    Bth stack sends notification related to the server.
    We receive connect notifications in this callback.
    
Arguments:
    Context - We receive server device context as the context
    Indication - Type of indication
    Parameters - Parameters of indication
--*/

{
	KdPrint(("\nPSCR! BthEchoSrvIndicationCallback: Enter\n"));
    switch(Indication)
    {
        case IndicationAddReference:
        case IndicationReleaseReference:
            break;
        case IndicationRemoteConnect:
        {
			
            PBTH_DEVICE_CONTEXT pDevContext = (PBTH_DEVICE_CONTEXT) Context;
			KdPrint(("PSCR! case IndicationRemoteConnect\n"));

            BthEchoSrvSendConnectResponse(pDevContext, Parameters);
            break;
        }
        case IndicationRemoteDisconnect:
        {
			KdPrint(("PSCR! case IndicationRemoteDisconnect\n"));
            ASSERT(FALSE);
            break;
        }
        case IndicationRemoteConfigRequest:
        case IndicationRemoteConfigResponse:
        case IndicationFreeExtraOptions:
            break;
    }
	KdPrint(("PSCR! BthEchoSrvIndicationCallback: Exit\n"));
}



__drv_maxIRQL(DISPATCH_LEVEL)
void
BthEchoSrvConnectionIndicationCallback(
    __in PVOID Context,
    __in INDICATION_CODE Indication,
    __in PINDICATION_PARAMETERS Parameters
    )
/*++
Description:
    Indication callback passed to bth stack while responding to connect.
    Bth stack sends notification related to the connection.
    We receive disconnect notifications in this callback
    
Arguments:
    Context - We receive connection object as the context
    Indication - Type of indication
    Parameters - Parameters of indication

--*/
{
    UNREFERENCED_PARAMETER(Parameters);

	KdPrint(("\nSCRD! BthEchoSrvConnectionIndicationCallback: Enter\n"));
   
    switch(Indication)
    {
        case IndicationAddReference:
            break;
        case IndicationReleaseReference:
            break;
        case IndicationRemoteConnect:
        {
            // We don't expect connect on this callback
			KdPrint(("SCRD! case IndicationRemoteConnect\n"));
            ASSERT(FALSE);
            break;
        }
        case IndicationRemoteDisconnect:
        {
			PBTH_DEVICE_CONTEXT pDevContext;
			PSMARTCARD_EXTENSION SmartcardExtension;
			
            WDFOBJECT connectionObject = (WDFOBJECT) Context;
            PBTHECHO_CONNECTION connection = GetConnectionObjectContext(connectionObject);
			KdPrint(("SCRD! case IndicationRemoteDisconnect\n"));

            BthEchoSrvDisconnectConnection(connection);

			pDevContext = GetServerDeviceContext(connection->Device);

		    SmartcardExtension = (PSMARTCARD_EXTENSION) &pDevContext->SmartCardExtension;
			//SmartcardExtension->ReaderExtension->StatusFileSelected = FALSE;
			CBUpdateCardState(SmartcardExtension, PSCR_ICC_ABSENT, FALSE);//CARD is present
            
            break;
        }
        case IndicationRemoteConfigRequest:
		{
			KdPrint(("SCRD! case IndicationRemoteConfigRequest\n"));
		}
        case IndicationRemoteConfigResponse:
        case IndicationFreeExtraOptions:
		{
			KdPrint(("SCRD! case IndicationFreeExtraOptions\n"));
			break;
		}
           
        default:
            ASSERT(FALSE);
    }
	KdPrint(("SCRD! BthEchoSrvConnectionIndicationCallback: Exit\n"));
}




__drv_maxIRQL(PASSIVE_LEVEL)
NTSTATUS
BthEchoSrvRegisterPSM(
    __in PBTH_DEVICE_CONTEXT pDevContext
    )
/*++
Description:    
	Registers Protocol/Service Multiplexer (PSM) that the channel uses
	to connect to the remote device.
--*/
{
    NTSTATUS status;
    struct _BRB_PSM * brb;
	KdPrint(("PSCR! BthEchoSrvRegisterPSM: Enter\n"));

    pDevContext->ProfileDrvInterface.BthReuseBrb(
        &(pDevContext->RegisterUnregisterBrb), 
        BRB_REGISTER_PSM
        );

    brb = (struct _BRB_PSM *) &(pDevContext->RegisterUnregisterBrb);
    
    // Send in our preferred PSM
    brb->Psm = pDevContext->Psm;

    status = BthEchoSharedSendBrbSynchronously(
        pDevContext->IoTarget,
        pDevContext->Request,
        (PBRB) brb,
        sizeof(*brb)
        );
    
    if (!NT_SUCCESS(status))
    {
		KdPrint( ("BRB_REGISTER_PSM failed, Status code %!STATUS!\n", status));

        return status;        
    }

    // Store PSM obtained
    pDevContext->Psm = brb->Psm;

	KdPrint(("PSCR! BthEchoSrvRegisterPSM: Exit\n"));

    return status;
}



__drv_maxIRQL(PASSIVE_LEVEL)
VOID
BthSrvUnregisterPSM(
    __in PBTH_DEVICE_CONTEXT pDevContext
    )
/*++
Description:    Unregisters server PSM.
--*/
{
    NTSTATUS status;
    struct _BRB_PSM * brb;

    PAGED_CODE();
	KdPrint(("PSCR! BthSrvUnregisterPSM: Enter\n"));

    if (pDevContext->Psm == 0)
        return;

	
	//Profile drivers use the BRB_UNREGISTER_PSM request to unregister
	//a Protocol/Service Multiplexer (PSM) that was previously registered

    pDevContext->ProfileDrvInterface.BthReuseBrb(
        &(pDevContext->RegisterUnregisterBrb), 
        BRB_UNREGISTER_PSM
        );

    brb = (struct _BRB_PSM *) &(pDevContext->RegisterUnregisterBrb);

    //
    // Format Brb
    //
    brb->Psm = pDevContext->Psm;

    status = BthEchoSharedSendBrbSynchronously(
        pDevContext->IoTarget,
        pDevContext->Request,
        (PBRB) brb,
        sizeof(*(brb))
        );

    if (!NT_SUCCESS(status))
    {
		KdPrint( ("BRB_UNREGISTER_PSM failed, Status code %x\n", status));

        // Send does not fail for resource reasons
        ASSERT(FALSE);      
    }
}



__drv_maxIRQL(PASSIVE_LEVEL)
NTSTATUS
BthEchoSrvRegisterL2CAPServer(
    __in PBTH_DEVICE_CONTEXT pDevContext
    )
/*++
Description: Registers L2CAP server. 
--*/
{
    NTSTATUS status;
    struct _BRB_L2CA_REGISTER_SERVER *brb;

	KdPrint(("PSCR! BthEchoSrvRegisterL2CAPServer: Enter\n"));
    pDevContext->ProfileDrvInterface.BthReuseBrb(
        &(pDevContext->RegisterUnregisterBrb), 
        BRB_L2CA_REGISTER_SERVER
        );

    brb = (struct _BRB_L2CA_REGISTER_SERVER *)
            &(pDevContext->RegisterUnregisterBrb);

    // Format brb
    brb->BtAddress = BTH_ADDR_NULL;
    brb->PSM = 0; //we have already registered the PSM
    //Indication callback passed to bth stack while registering server
	brb->IndicationCallback = &BthEchoSrvIndicationCallback;
    brb->IndicationCallbackContext = pDevContext;
    brb->IndicationFlags = 0;
    brb->ReferenceObject = WdfDeviceWdmGetDeviceObject(pDevContext->Device);

    status = BthEchoSharedSendBrbSynchronously(
        pDevContext->IoTarget,
        pDevContext->Request,
        (PBRB) brb,
        sizeof(*brb)
        );

    if (!NT_SUCCESS(status))
    {
		KdPrint( ("BRB_L2CA_REGISTER_SERVER failed, Status code %x\n", status));
        return status;        
    }

    // Store server handle
    pDevContext->L2CAPServerHandle = brb->ServerHandle;

	KdPrint(("PSCR! BthEchoSrvRegisterL2CAPServer: Exit\n"));

    return status;
}



__drv_maxIRQL(PASSIVE_LEVEL)
VOID
BthSrvUnregisterL2CAPServer(
    __in PBTH_DEVICE_CONTEXT pDevContext
    )
{
    NTSTATUS status;
    struct _BRB_L2CA_UNREGISTER_SERVER *brb;

    PAGED_CODE();

	KdPrint(("PSCR! BthSrvUnregisterL2CAPServer: Enter\n"));
    if (NULL == pDevContext->L2CAPServerHandle)
        return;

    pDevContext->ProfileDrvInterface.BthReuseBrb(
        &(pDevContext->RegisterUnregisterBrb), 
        BRB_L2CA_UNREGISTER_SERVER
        );

    brb = (struct _BRB_L2CA_UNREGISTER_SERVER *)
            &(pDevContext->RegisterUnregisterBrb);

    //
    // Format Brb
    //
    brb->BtAddress = BTH_ADDR_NULL ;//DevCtx->LocalAddress;
    brb->Psm = 0; //since we will use unregister PSM to unregister.
    brb->ServerHandle = pDevContext->L2CAPServerHandle;

    status = BthEchoSharedSendBrbSynchronously(
        pDevContext->IoTarget,
        pDevContext->Request,
        (PBRB) brb,
        sizeof(*(brb))
        );

    if (!NT_SUCCESS(status))
    {
		KdPrint( ("BRB_L2CA_UNREGISTER_SERVER failed, Status code %!STATUS!\n", status));

        // Send does not fail for resource reasons
        ASSERT(FALSE);
        
        return;        
    }

    pDevContext->L2CAPServerHandle = NULL;
}



__drv_maxIRQL(PASSIVE_LEVEL)
NTSTATUS
BthEchoSrvPublishSdpRecord(
    __in PBTH_DEVICE_CONTEXT pDevContext,
    __in PUCHAR SdpRecord,
    __in ULONG  SdpRecordLength
    )
/*++
Description:
    Publishes server SDP record. The SDP record built using CreateSdpRecord
	in sdp.c is published by this function

Arguments:
    pDevContext - Device context of the server
    SdpRecord - Sdp record to publish
    SdpRecordLength - Sdp record length
--*/
{
    NTSTATUS status, statusReuse;
    WDF_MEMORY_DESCRIPTOR inMemDesc;
    WDF_MEMORY_DESCRIPTOR outMemDesc;
    WDF_REQUEST_REUSE_PARAMS ReuseParams;
    HANDLE_SDP sdpRecordHandle;

    PAGED_CODE();
	KdPrint(("SCRD! BthEchoSrvPublishSdpRecord: Enter\n"));
    
    WDF_REQUEST_REUSE_PARAMS_INIT(&ReuseParams, WDF_REQUEST_REUSE_NO_FLAGS, STATUS_NOT_SUPPORTED);
    statusReuse = WdfRequestReuse(pDevContext->Request, &ReuseParams);    
    ASSERT(NT_SUCCESS(statusReuse));
    UNREFERENCED_PARAMETER(statusReuse);

    WDF_MEMORY_DESCRIPTOR_INIT_BUFFER(
        &inMemDesc,
        SdpRecord,
        SdpRecordLength
        );

    RtlZeroMemory( &sdpRecordHandle, sizeof(HANDLE_SDP) );
    
    WDF_MEMORY_DESCRIPTOR_INIT_BUFFER(
        &outMemDesc,
        &sdpRecordHandle,
        sizeof(HANDLE_SDP)
        );

	/*
	The IOCTL_BTH_SDP_SUBMIT_RECORD request allows a profile driver to add an SDP record
	to the local SDP server, allowing the client to advertise that a service is available
	on the local computer. The profile driver calls IOCTL_BTH_SDP_REMOVE_RECORD to stop
	advertising the server on the local SDP server. */
    //builds a device control request and sends it synchronously to the I/O target
	status = WdfIoTargetSendIoctlSynchronously(
        pDevContext->IoTarget,
        pDevContext->Request,
        IOCTL_BTH_SDP_SUBMIT_RECORD,
        &inMemDesc,
        &outMemDesc,
        NULL,   //sendOptions
        NULL    //bytesReturned
        );

    if (!NT_SUCCESS(status))
    {
		KdPrint( ("PSCR! BthEchoSrvPublishSdpRecord: IOCTL_BTH_SDP_SUBMIT_RECORD failed, Status code %!STATUS!\n", status));

        return status;
    }

    pDevContext->SdpRecordHandle = sdpRecordHandle;

	KdPrint(("SCRD! BthEchoSrvPublishSdpRecord: Exit\n"));

    return status;    
}




__drv_maxIRQL(PASSIVE_LEVEL)
VOID
BthSrvRemoveSdpRecord(
    __in PBTH_DEVICE_CONTEXT pDevContext
    )
{
    NTSTATUS status, statusReuse;
    WDF_MEMORY_DESCRIPTOR inMemDesc;
    WDF_REQUEST_REUSE_PARAMS ReuseParams;
    HANDLE_SDP sdpRecordHandle;

    PAGED_CODE();

	KdPrint(("PSCR! BthSrvRemoveSdpRecord: Enter\n"));

	/*A driver must not try to reuse an I/O request object until the previous
	request has completed. For this reason, many drivers call the WdfRequestReuse
	method to reinitialize the I/O request object from their I/O completion callbacks
	*/  
    WDF_REQUEST_REUSE_PARAMS_INIT(&ReuseParams, WDF_REQUEST_REUSE_NO_FLAGS, STATUS_NOT_SUPPORTED);
    statusReuse =  WdfRequestReuse(pDevContext->Request, &ReuseParams);    
    ASSERT(NT_SUCCESS(statusReuse));
    UNREFERENCED_PARAMETER(statusReuse);//to avoid compiler warnings about unreferenced parameters

    sdpRecordHandle = pDevContext->SdpRecordHandle;

    WDF_MEMORY_DESCRIPTOR_INIT_BUFFER(
        &inMemDesc,
        &sdpRecordHandle,
        sizeof(HANDLE_SDP)
        );
    
	//The IOCTL_BTH_SDP_REMOVE_RECORD request removes a local SDP record
	//that the profile driver previously submitted
    status = WdfIoTargetSendIoctlSynchronously(
        pDevContext->IoTarget,
        pDevContext->Request,
        IOCTL_BTH_SDP_REMOVE_RECORD,
        &inMemDesc,
        NULL,   //outMemDesc
        NULL,   //sendOptions
        NULL    //bytesReturned
        );

    if (!NT_SUCCESS(status))
    {
		KdPrint( ("IOCTL_BTH_SDP_REMOVE_RECORD failed, Status code %s\n", status));

        // Send does not fail for resource reasons
        ASSERT(FALSE); //macro: If the expression is false, it breaks into the kernel debugger

        return;
    }

    pDevContext->SdpRecordHandle = HANDLE_SDP_NULL;        
}




void
BthEchoSrvRemoteConnectResponseCompletion(
    __in WDFREQUEST  Request,
    __in WDFIOTARGET  Target,
    __in PWDF_REQUEST_COMPLETION_PARAMS  Params,
    __in WDFCONTEXT  Context
    )
/*++
Description:
    Completion routine for BRB_L2CA_OPEN_CHANNEL_RESPONSE BRB sent in BthEchoSrvSendConnectResponse.

    If the request completed successfully we call BthEchoSrvConnectionStateConnected
    function implemented in device.c. This function initializes and submits
    continuous readers to read the data coming from client which then is used
    for echo.

Arguments:
    Request - Request that completed
    Target - Target to which request was sent
    Params - Request completion parameters
    Context - We receive Brb as the context. This Brb is connection->ConnectDisconnectBrb,
	and is not allocated separately hence it is not freed here.
--*/
{
    NTSTATUS status;
	
    struct _BRB_L2CA_OPEN_CHANNEL *brb;    
    PBTHECHO_CONNECTION connection;
    WDFOBJECT connectionObject;
	PBTH_DEVICE_CONTEXT pDevContext;
	PSMARTCARD_EXTENSION SmartcardExtension;

    UNREFERENCED_PARAMETER(Target);
    UNREFERENCED_PARAMETER(Request); //we reuse the request, hence it is not needed here
    
    status = Params->IoStatus.Status;

	KdPrint(("\nSCRD! ====>>  BthEchoSrvRemoteConnectResponseCompletion: Enter, IoStatus.status =  %s\n", status));


	 
    brb = (struct _BRB_L2CA_OPEN_CHANNEL *) Context;
	pDevContext = GetServerDeviceContext(WdfWdmDeviceGetWdfDeviceHandle(brb->ReferenceObject));

    //
    // We receice connection object as the context in the BRB
    //
    connectionObject = (WDFOBJECT) brb->Hdr.ClientContext[0];
    connection = GetConnectionObjectContext(connectionObject);
    
    if(NT_SUCCESS(status))
    {
        connection->OutMTU = brb->OutResults.Params.Mtu;
        connection->InMTU = brb->InResults.Params.Mtu;
        connection->ChannelHandle = brb->ChannelHandle;
        connection->RemoteAddress = brb->BtAddress;

        //
        // Check if we already received a disconnect request and if so disconnect
        //
        WdfSpinLockAcquire(connection->ConnectionLock);

        if (connection->ConnectionState == ConnectionStateDisconnecting) 
        {
            //
            // We allow transition to disconnected state only from connected state
            //
            // Set the state to Connected and call BthEchoConnectionObjectRemoteDisconnect
            //
            connection->ConnectionState = ConnectionStateConnected;
            WdfSpinLockRelease(connection->ConnectionLock);

			KdPrint(("SCRD! BthEchoSrvRemoteConnectResponseCompletion: disconnect has been received for connection 0x%p\n", connection));
    
            BthEchoConnectionObjectRemoteDisconnect(connection->IoTarget, &connection->ProfileDrvInterface, connection);
        }
        else
        {
            connection->ConnectionState = ConnectionStateConnected;
            WdfSpinLockRelease(connection->ConnectionLock);

			KdPrint(("SCRD! BthEchoSrvRemoteConnectResponseCompletion: Connection completed, connection: 0x%p\n", connection));

            //
            // Call BthEchoSrvConnectionStateConnected to perform post connect processing:
            // 1. initialize and submit continuous readers to read the data coming from client
			// 2. change card state to present and inform the system
            //
            status = BthEchoSrvConnectionStateConnected(connectionObject);
            if (!NT_SUCCESS(status))    
            {
                //
                // If the post connect processing failed, let us disconnect
                //
                BthEchoConnectionObjectRemoteDisconnect(connection->IoTarget, &connection->ProfileDrvInterface, connection);
            } 
			else
			{
				//
				//if everything is ok then change card state to SCARD_PRESENT
				//
			    SmartcardExtension = (PSMARTCARD_EXTENSION) &pDevContext->SmartCardExtension;
				//SmartcardExtension->ReaderExtension->StatusFileSelected = FALSE;
				CBUpdateCardState(SmartcardExtension, PSCR_ICC_PRESENT, FALSE);//CARD is present
			}


        }
    }
    else
    {
        BOOLEAN setDisconnectEvent = FALSE;
        
        WdfSpinLockAcquire(connection->ConnectionLock);
        if (connection->ConnectionState == ConnectionStateDisconnecting) 
        {
            setDisconnectEvent = TRUE;
        }

        connection->ConnectionState = ConnectionStateConnectFailed;
        WdfSpinLockRelease(connection->ConnectionLock);

        if (setDisconnectEvent)
			KeSetEvent(&connection->DisconnectEvent, 0, FALSE);
        
    }

    if (!NT_SUCCESS(status))
    {        
		WdfSpinLockAcquire(connection->ConnectionLock);
		RemoveEntryList(&connection->ConnectionListEntry);
		WdfSpinLockRelease(connection->ConnectionLock);
        
        WdfObjectDelete(connectionObject);        
    }

	KdPrint(("SCRD! <<====  BthEchoSrvRemoteConnectResponseCompletion: Exit\n"));
    return;    
}



__drv_maxIRQL(DISPATCH_LEVEL)
NTSTATUS
BthEchoSrvSendConnectResponse(
    __in PBTH_DEVICE_CONTEXT pDevContext,
    __in PINDICATION_PARAMETERS ConnectParams
    )
/*++
Description:
    Respond to connect indication received. This function is invoked by
	BthEchoSrvIndicationCallback when connect indication is received.

Arguments:
    pDevContext - Server device context
    ConnectParams - Connect indication parameters
--*/
{
    NTSTATUS statusReuse, status = STATUS_SUCCESS;
    WDF_REQUEST_REUSE_PARAMS reuseParams;
    struct _BRB_L2CA_OPEN_CHANNEL *brb = NULL;
    WDFOBJECT connectionObject = NULL;
    PBTHECHO_CONNECTION connection = NULL;
    
    KdPrint(("SCRD!   BthEchoSrvSendConnectResponse: Enter\n"));    


    //
    // We create the connection object as the first step so that if we receive 
    // remove before connect response is completed, we can wait for connection and disconnect.
    //
    status = BthEchoConnectionObjectCreate(
        pDevContext,
        pDevContext->Device,
        &connectionObject
        );

    if (!NT_SUCCESS(status))
    {
        goto exit;
    }

    connection = GetConnectionObjectContext(connectionObject);

    connection->ConnectionState = ConnectionStateConnecting;

    //
    // Insert the connection object in the conenction list that we track
    //
    WdfSpinLockAcquire(pDevContext->ConnectionListLock);
    InsertTailList(&pDevContext->ConnectionList, &connection->ConnectionListEntry);
    WdfSpinLockRelease(pDevContext->ConnectionListLock);    

    WDF_REQUEST_REUSE_PARAMS_INIT(&reuseParams, WDF_REQUEST_REUSE_NO_FLAGS, STATUS_NOT_SUPPORTED);
    statusReuse = WdfRequestReuse(connection->ConnectDisconnectRequest, &reuseParams);
    ASSERT(NT_SUCCESS(statusReuse));
    UNREFERENCED_PARAMETER(statusReuse);
    
    brb = (struct _BRB_L2CA_OPEN_CHANNEL*) &(connection->ConnectDisconnectBrb);
    pDevContext->ProfileDrvInterface.BthReuseBrb((PBRB)brb, BRB_L2CA_OPEN_CHANNEL_RESPONSE);

    brb->Hdr.ClientContext[0] = connectionObject;
    brb->BtAddress = ConnectParams->BtAddress;
    brb->Psm = ConnectParams->Parameters.Connect.Request.PSM;
    brb->ChannelHandle = ConnectParams->ConnectionHandle;
    brb->Response = CONNECT_RSP_RESULT_SUCCESS;

    //brb->ChannelFlags = CF_ROLE_EITHER;
	brb->ChannelFlags = CF_LINK_ENCRYPTED;


    //brb->ConfigOut.Flags = CF_LINK_AUTHENTICATED | CF_LINK_ENCRYPTED;
    brb->ConfigOut.Flags = 0;
    //brb->ConfigIn.Flags = CF_LINK_AUTHENTICATED | CF_LINK_ENCRYPTED;
	brb->ConfigIn.Flags = 0;

    brb->ConfigOut.Flags |= CFG_MTU;
    brb->ConfigOut.Mtu.Max = L2CAP_DEFAULT_MTU;
    brb->ConfigOut.Mtu.Min = L2CAP_MIN_MTU;
    brb->ConfigOut.Mtu.Preferred = L2CAP_DEFAULT_MTU;

    brb->ConfigIn.Flags = CFG_MTU;
    brb->ConfigIn.Mtu.Max = brb->ConfigOut.Mtu.Max;
    brb->ConfigIn.Mtu.Min = brb->ConfigOut.Mtu.Min;
    brb->ConfigIn.Mtu.Preferred = brb->ConfigOut.Mtu.Max;

    //
    // Get notifications about disconnect
    //
    brb->CallbackFlags = CALLBACK_DISCONNECT;
    brb->Callback = &BthEchoSrvConnectionIndicationCallback;
    brb->CallbackContext = connectionObject;
    brb->ReferenceObject = (PVOID) WdfDeviceWdmGetDeviceObject(pDevContext->Device);

    status = BthEchoSharedSendBrbAsync(
        pDevContext->IoTarget,
        connection->ConnectDisconnectRequest,
        (PBRB) brb,
        sizeof(*brb),
        BthEchoSrvRemoteConnectResponseCompletion,
        brb //context
        );

    if (!NT_SUCCESS(status))
    {
		KdPrint(("!!!!!   BthEchoSrvSendConnectResponse: BthEchoSharedSendBrbAsync failed, status = (%lx)\n", status));
    }
    
exit:    
    
    if(!NT_SUCCESS(status) && connectionObject)
    {
        //
        // If we failed to connect remove connectionfrom list and delete the object
        //

        //
        // connection should not be NULL if connectionObject is not NULL
        // since first thing we do after creating connectionObject is to
        // get context which gives us connection
        //
        ASSERT(connection != NULL);

        if (connection != NULL)
        {
            connection->ConnectionState = ConnectionStateConnectFailed;

		    WdfSpinLockAcquire(connection->ConnectionLock);
			RemoveEntryList(&connection->ConnectionListEntry);
			WdfSpinLockRelease(connection->ConnectionLock); 
        }

        WdfObjectDelete(connectionObject);
    }

	KdPrint(("SCRD!   BthEchoSrvSendConnectResponse: Exit, status = (%lx)\n", status));

    return status;
}



__drv_maxIRQL(DISPATCH_LEVEL)
VOID
BthEchoSrvDisconnectConnection(
    __in PBTHECHO_CONNECTION Connection
    )    
/*++
Description:
    Disconnects connection, removes it from the list and deletes the connection
    object.    
    This helper function can be called for disconnect only after connect is
    completed.
--*/
{
    WDFOBJECT connectionObject;

    //
    // If BthEchoConnectionObjectRemoteDisconnect returns FALSE
    // connection is already disconnected
    // 
	KdPrint(("PSCR! BthEchoSrvDisconnectConnection: Enter\n"));

    if (BthEchoConnectionObjectRemoteDisconnect(Connection->IoTarget, &Connection->ProfileDrvInterface, Connection))
    {
	    WdfSpinLockAcquire(Connection->ConnectionLock);
		RemoveEntryList(&Connection->ConnectionListEntry);
		WdfSpinLockRelease(Connection->ConnectionLock); 

        connectionObject = WdfObjectContextGetObject(Connection);

        WdfObjectDelete(connectionObject);    
    }
	KdPrint(("PSCR! BthEchoSrvDisconnectConnection: Exit\n"));
}



__drv_maxIRQL(DISPATCH_LEVEL)
VOID
BthEchoSrvDisconnectConnectionsOnRemove(
    __in PBTH_DEVICE_CONTEXT pDevContext
    )
/*++
Description:
    This routine is invoked by BthEchoSrvEvtDeviceSelfManagedIoCleanup
    during remove. We disconnect all the outstanding connections
    in this routine.

    This routine does not wait for completion of the disconnect. connection object's
    cleanup callback waits for that.
--*/
{
	
    BOOLEAN True = TRUE;

	KdPrint(("PSCR! BthEchoSrvDisconnectConnectionsOnRemove: Enter\n"));


    //
    // We will not get any new connections beacuse we have unregistered server
    // at this point.
    //
    // Although it is possible that connections from the current list 
    // get disconnected. In such case the following things protect us:
    //     1. Our reference on connection object keeps connection object alive
    //     2. Connection state tracking ensures that disconnect is sent
    //        only once
    //

    while (True)
    {
        PLIST_ENTRY ple;
        PBTHECHO_CONNECTION connection;
        
        WdfSpinLockAcquire(pDevContext->ConnectionListLock);       

        if (IsListEmpty(&pDevContext->ConnectionList))
        {
            WdfSpinLockRelease(pDevContext->ConnectionListLock);       
            break;            
        }

        ple = pDevContext->ConnectionList.Flink;

        RemoveEntryList(ple);

        connection = CONTAINING_RECORD(ple, BTHECHO_CONNECTION, ConnectionListEntry);

        //
        // Add a reference since we will be touching connection object
        // outside of the connection list lock
        //
        WdfObjectReference(WdfObjectContextGetObject(connection));        
        
        WdfSpinLockRelease(pDevContext->ConnectionListLock);       

        //
        // Invoke disconnect outside the lock
        // If client disconnects in the meantime connection state tracking ensures
        // that we send disconnect only once
        //
        BthEchoConnectionObjectRemoteDisconnect(pDevContext->IoTarget, &pDevContext->ProfileDrvInterface ,connection);

        //
        // Release the reference added above
        //
        WdfObjectDereference(WdfObjectContextGetObject(connection));        

        //
        // We don't explicitly delete connection object as it is a child of device object
        // which will shortly be deleted.
        //
    }

	KdPrint(("PSCR! BthEchoSrvDisconnectConnectionsOnRemove: Exit\n"));
}    



__drv_maxIRQL(DISPATCH_LEVEL)
NTSTATUS
BthEchoSrvConnectionStateConnected(
    WDFOBJECT ConnectionObject
    )
/*++
Routine Description:
    This routine is invoked by BthEchoSrvRemoteConnectResponseCompletion
    when connect response is completed.
    We initialize and submit continous readers in this routine.

Arguments:
    ConnectionObject - Connection object for which connect response completed
--*/
{
    PBTHECHO_CONNECTION connection;
    NTSTATUS status = STATUS_SUCCESS;

	KdPrint(("SCRD! -> BthEchoSrvConnectionStateConnected: Enter\n"));


    connection = GetConnectionObjectContext(ConnectionObject);

    status = BthEchoConnectionObjectInitializeContinuousReader(
        connection,
        BthEchoSrvConnectionObjectContReaderReadCompletedCallback,
        BthEchoSrvConnectionObjectContReaderFailedCallback,
        BthEchoSampleMaxDataLength
        );

    if (!NT_SUCCESS(status))
        return status;

    status = BthEchoConnectionObjectContinuousReaderSubmitReaders( connection );

	KdPrint(("SCRD! <- BthEchoSrvConnectionStateConnected: Exit,  status = (%lx)\n", status));
 
    return status;
}




__drv_maxIRQL(DISPATCH_LEVEL)
VOID
BthEchoSrvConnectionObjectContReaderReadCompletedCallback(
    __in PBTHECHO_CONNECTION Connection,
    __in PVOID Buffer,
    __in size_t BufferLength
    )
/*++
Routine Description:
    This routine is invoked by the continous reader in BthEchoRepeatReaderPendingReadCompletion
	when read completes. Routine submits response to smclib

Arguments:
    Connection - Connection whose continous reader had read completion
    Bufer - Buffer which received read
    SrcBufferLength - Length of read

--*/
{
	NTSTATUS NTStatus = STATUS_SUCCESS;
	PBTH_DEVICE_CONTEXT pDevContext;
	PSMARTCARD_EXTENSION SmartcardExtension;
	ULONG i;

	KdPrint(("SCRD! BthEchoSrvConnectionObjectContReaderReadCompletedCallback: Enter\n"));

	pDevContext = GetServerDeviceContext(Connection->Device);
	SmartcardExtension = &pDevContext->SmartCardExtension;


	// check if response waiting time has run out (server is not waiting for response)
	if(pDevContext->WaitingResponse == FALSE)
	return;

    SmartcardExtension->SmartcardReply.BufferLength = BufferLength;

    SysCopyMemory(
                 SmartcardExtension->SmartcardReply.Buffer,
                 Buffer,
                 BufferLength
                 );


    // copies data from the receive buffer to the user buffer and checks the T=0 protocol status
   // NTStatus = SmartcardT0Reply( SmartcardExtension );

	KeSetEvent(
		&pDevContext->ReceivedResponseEvt,
		(KPRIORITY) 0,
		FALSE)
		;
	KdPrint(("  ----   \n" ));
	for (i = 0; i< BufferLength; i++)
		KdPrint(("  SCRD! CBT0Receive: Response [%d] = 0x%x\n", i, SmartcardExtension->SmartcardReply.Buffer[i] ));

	KdPrint(("SCRD! BthEchoSrvConnectionObjectContReaderReadCompletedCallback: Exit (%lx)\n", NTStatus));
}




__drv_maxIRQL(DISPATCH_LEVEL)
VOID
BthEchoSrvConnectionObjectContReaderFailedCallback(
    __in PBTHECHO_CONNECTION Connection
    )
/*++
Routine Description:

    This routine is invoked by the continous reader when there is a failure
    in read submission on completion

    In response we disconnect the remote connection

Arguments:

    DevCtxHdr - Device context
    Connection - Connection whose continous reader had read completion
    Bufer - Buffer which received read
    SrcBufferLength - Length of read

--*/
{    
	KdPrint(("SCRD! BthEchoSrvConnectionObjectContReaderFailedCallback: Enter\n"));
    BthEchoSrvDisconnectConnection(Connection);
	KdPrint(("SCRD! BthEchoSrvConnectionObjectContReaderFailedCallback: Exit\n"));
}


__drv_maxIRQL(PASSIVE_LEVEL)
NTSTATUS
BthEchoSharedRetrieveLocalBthAddress(
    __in PBTH_DEVICE_CONTEXT pDevContext
    )
/*++
Description:
    Retrieves the local bth address (burnt into device => doesn't change).
--*/
{
    NTSTATUS status = STATUS_SUCCESS;
    struct _BRB_GET_LOCAL_BD_ADDR * brb = NULL;
    
	KdPrint(("SCRD! BthEchoSharedRetrieveLocalBthAddress: Enter\n"));

	// allocates a Bluetooth request block (BRB) of the BRB_HCI_GET_LOCAL_BD_ADDR type
    brb = (struct _BRB_GET_LOCAL_BD_ADDR *)
		pDevContext->ProfileDrvInterface.BthAllocateBrb(
            BRB_HCI_GET_LOCAL_BD_ADDR, 
            POOLTAG_BTHECHOSAMPLE
            );

    if(brb == NULL)
    {
        status = STATUS_INSUFFICIENT_RESOURCES;

		KdPrint( ("Failed to allocate brb BRB_HCI_GET_LOCAL_BD_ADDR, returning status code %s\n", status));

        return status;
    }

    status = BthEchoSharedSendBrbSynchronously(
        pDevContext->IoTarget,
        pDevContext->Request,
        (PBRB) brb,
        sizeof(*brb)
        );

    if (!NT_SUCCESS(status))
    { 
		KdPrint( ("Retrieving local bth address failed, Status code %s\n", status));

        pDevContext->ProfileDrvInterface.BthFreeBrb((PBRB)brb);
		return status;
    }

    pDevContext->LocalBthAddr = brb->BtAddress;

	KdPrint(("SCRD! BthEchoSharedRetrieveLocalBthAddress: Exit\n"));

    return status;
}



void
BthEchoSrvWriteCompletion(
    __in WDFREQUEST  Request,
    __in WDFIOTARGET  Target,
    __in PWDF_REQUEST_COMPLETION_PARAMS  Params,
    __in WDFCONTEXT  Context
    )
/*++
Description:
    Completion routine for echo L2Ca transfer

Arguments:
    Request - Request that completed
    Target - Target to which request was sent
    Params - Request completion parameters
    Context - We receive connection as the context

--*/
{
    NTSTATUS status;

    UNREFERENCED_PARAMETER(Target);
    UNREFERENCED_PARAMETER(Context);

    status = Params->IoStatus.Status;
    
	KdPrint( ("SCRD! BthEchoSrvWriteCompletion: Enter, status = (%lx)\n", status));

    WdfObjectDelete(Request);    

    /*
    We don't attempt to disconnect in case of failure here because we expect
	the failure only be due to device/channel disconnect, in which case disconnect
    would happen anyway.
	
	Without taking a reference on the connection object while submitting echo write
    we cannot touch connection object here as connection could get disconnected
	and connection object could get freed.
    */
}




__drv_maxIRQL(DISPATCH_LEVEL)
VOID
BthEchoSrvSendEcho(
	__in WDFDEVICE Device,
	__in WDFIOTARGET IoTarget,
    __in PBTHECHO_CONNECTION Connection,
    __in PVOID SrcBuffer,
    __in size_t SrcBufferLength
    )
/*++
Routine Description:
    Performs L2Cap transfer to client 
    This routine is invoked by one the callback functions registered with Smclib


Arguments:
    DevCtxHdr - Device context
    Connection - Connection whose continous reader had read completion
    SrcBuffer - Source buffer
    SrcBufferLength - Length of the source buffer

--*/
{
    NTSTATUS status;
    WDF_OBJECT_ATTRIBUTES attributes;
    WDFREQUEST request;
    WDFMEMORY memory;
    PVOID dataBuffer;
    struct _BRB_L2CA_ACL_TRANSFER *brb = NULL;

	KdPrint(("SCRD! BthEchoSrvSendEcho: Enter\n"));

    WDF_OBJECT_ATTRIBUTES_INIT(&attributes);
    attributes.ParentObject = Device;

    WDF_OBJECT_ATTRIBUTES_INIT_CONTEXT_TYPE(&attributes, BRB);

    status = WdfRequestCreate(
        &attributes,
        IoTarget,
        &request
        );                    

    if (!NT_SUCCESS(status))
    {
		KdPrint( ("Creating request for echo failed, Status code %s\n", status));

        goto exit;
    }

    WDF_OBJECT_ATTRIBUTES_INIT(&attributes);
    attributes.ParentObject = request;

    status = WdfMemoryCreate(
        &attributes,
        NonPagedPool,
        POOLTAG_BTHECHOSAMPLE,
        SrcBufferLength,
        &memory,
        &dataBuffer
        );

    if (!NT_SUCCESS(status))
    {
		KdPrint( ("Creating memory for pending read failed, Status code %s\n", status));

        goto exit;
    }

    memcpy(dataBuffer, SrcBuffer, SrcBufferLength);

    brb = (struct _BRB_L2CA_ACL_TRANSFER *)GetEchoRequestContext(request);

    status = BthEchoConnectionObjectFormatRequestForL2CaTransfer(
        Connection,
        request,
        &brb,
        memory,
        ACL_TRANSFER_DIRECTION_OUT
        );

    if (!NT_SUCCESS(status))
    {
        goto exit;
    }

    //
    // Set a CompletionRoutine callback function.
    //
    
    WdfRequestSetCompletionRoutine(
        request,
        BthEchoSrvWriteCompletion,
        Connection
        );

    if (FALSE == WdfRequestSend(
        request,
        IoTarget,
        NULL
        ))
    {
        status = WdfRequestGetStatus(request);

		KdPrint( ("Request send failed for request 0x%p, Brb 0x%p, Status code %s\n",
			request, brb, status));

        goto exit;
    }    

exit:
    if (!NT_SUCCESS(status))
    {
        WdfObjectDelete(request);

        //
        // If we failed disconnect
        //
        BthEchoSrvDisconnectConnection(Connection);        
    } 
	KdPrint(("SCRD! BthEchoSrvSendEcho: Exit\n"));
}