/*++

Copyright (c) Microsoft Corporation.  All rights reserved.

    THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
    KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
    IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
    PURPOSE.

Module Name:
    Serverblabla.h

Abstract:
    Contains declarations for Bluetooth server functionality.

--*/


#include "guid.h"
#include "connection.h"


#define POOLTAG_BTHECHOSAMPLE 'htbw'

__drv_maxIRQL(PASSIVE_LEVEL)
NTSTATUS
BthEchoSrvRegisterPSM(
    __in PBTH_DEVICE_CONTEXT DevCtx
    );

__drv_maxIRQL(PASSIVE_LEVEL)
VOID
BthSrvUnregisterPSM(
    __in PBTH_DEVICE_CONTEXT DevCtx
    );

__drv_maxIRQL(PASSIVE_LEVEL)
NTSTATUS
BthEchoSrvRegisterL2CAPServer(
    __in PBTH_DEVICE_CONTEXT DevCtx
    );

__drv_maxIRQL(PASSIVE_LEVEL)
VOID
BthSrvUnregisterL2CAPServer(
    __in PBTH_DEVICE_CONTEXT DevCtx
    );

__drv_maxIRQL(PASSIVE_LEVEL)
NTSTATUS
BthEchoSrvPublishSdpRecord(
    __in PBTH_DEVICE_CONTEXT pDevContext,
    __in PUCHAR SdpRecord,
    __in ULONG  SdpRecordLength
    );

__drv_maxIRQL(PASSIVE_LEVEL)
VOID
BthSrvRemoveSdpRecord(
    __in PBTH_DEVICE_CONTEXT DevCtx
    );

__drv_maxIRQL(DISPATCH_LEVEL)
NTSTATUS
BthEchoSrvSendConnectResponse(
    __in PBTH_DEVICE_CONTEXT DevCtx,
    __in PINDICATION_PARAMETERS ConnectParams
    );

__drv_maxIRQL(DISPATCH_LEVEL)
VOID
BthEchoSrvDisconnectConnection(
    __in PBTHECHO_CONNECTION Connection
    );

__drv_maxIRQL(DISPATCH_LEVEL)
VOID
BthEchoSrvDisconnectConnectionsOnRemove(
    __in PBTH_DEVICE_CONTEXT devCtx
    );


EVT_WDF_REQUEST_COMPLETION_ROUTINE BthEchoSrvRemoteConnectResponseCompletion;



//////////////////////////////////////////////////////
// Device specific functionality invoked by server.c
//////////////////////////////////////////////////////

__drv_maxIRQL(DISPATCH_LEVEL)
NTSTATUS
BthEchoSrvConnectionStateConnected(
    WDFOBJECT ConnectionObject
    );

////////////////////////////////////////////////////////////////////
// Continuous reader callbacks (invoked by connection.c)
////////////////////////////////////////////////////////////////////

__drv_maxIRQL(DISPATCH_LEVEL)
VOID
BthEchoSrvConnectionObjectContReaderReadCompletedCallback(
    __in PBTHECHO_CONNECTION Connection,
    __in PVOID Buffer,
    __in size_t BufferLength
    );

__drv_maxIRQL(DISPATCH_LEVEL)
VOID
BthEchoSrvConnectionObjectContReaderFailedCallback(
    __in PBTHECHO_CONNECTION Connection
    );


///////////////////////////////////////////////////////////////////////////
//
/////////////////////////

__drv_maxIRQL(DISPATCH_LEVEL)
VOID
BthEchoSrvSendEcho(
   	__in WDFDEVICE Device,
	__in WDFIOTARGET IoTarget,
    __in PBTHECHO_CONNECTION Connection,
    __in PVOID SrcBuffer,
    __in size_t SrcBufferLength
    );

WDF_DECLARE_CONTEXT_TYPE_WITH_NAME(BRB, GetEchoRequestContext)    

EVT_WDF_REQUEST_COMPLETION_ROUTINE BthEchoSrvWriteCompletion; 



///////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////

enum _BTHECHOSAMPLE_MAX_DATA_LENGTH
{
    BthEchoSampleMaxDataLength = 256
};


__drv_maxIRQL(PASSIVE_LEVEL)
NTSTATUS
BthEchoSharedRetrieveLocalBthAddress(
    __in PBTH_DEVICE_CONTEXT pDevContext
    );
