

/*
 * Copyright (c) 2010 NTNU/KTH author Oleksandr Bodriagov. All rights reserved. :)
 */

/*
 *  TOTPClient.java	 05/03/2010

 *  This class acts as a client for JCtotp applet. It gets the PIN
 *  and the current time and transfers them to the applet for HMAC-based
 *  one-time password generation. The applet returns a 32 bit byte array
 *  that is converted into a 6-digit number (password) and shown on the screen.
 */

package totp;

import java.io.*;
import javax.microedition.midlet.MIDlet;
import javax.microedition.lcdui.*;
import javax.microedition.apdu.APDUConnection;
import javax.microedition.io.Connector;
import org.w3c.dom.*;
import org.w3c.dom.svg.*;
import javax.microedition.m2g.*;
import totp.CardCommunicator;



public class TOTPClient extends MIDlet implements CommandListener, ItemCommandListener, Runnable //Runnable for starting new threads
 {
     final static String svgname= "resultForm.svg";
     private final String appletAID = "apdu:0;target=A0.00.00.00.62.91";
     final static int totpNumberDigits = 6;//number of digits in the OTP
     private APDUConnection cardConnection;
     private CardCommunicator cardCommunicator;

     //--------GUI--------------------------------------------------
     private Display mdisplay;
     private Displayable msvgForm;

     private Form mForm;
     private Form waitForm;
     private Form errorForm;
     private Form hotpForm;

     private Command generateCommand;
     private Command exitCommand;
     private Command backCommand;
     private Command cancelCommand;

     private TextField pinField;
     private StringItem generate;
     private StringItem hotpLine;
     private StringItem errLine;
     private Spacer mSpace1;
     private Spacer mSpace2;

     private SVGElement svgHotp;
     private SVGAnimator svgAnime;



     // Constructor. creates GUI
     public TOTPClient()
     {
         cardCommunicator = new CardCommunicator();

         //forms used by the MIDlet
         mForm = new Form("One-time password");
         waitForm = new Form("One-time password");
         errorForm = new Form("One-time password");
         hotpForm = new Form("One-time password");

         exitCommand = new Command("Exit", Command.EXIT, 0);
         cancelCommand = new Command("Cancel", Command.CANCEL, 0);
         backCommand = new Command("Back", Command.BACK, 0);

         //mForm
         generate = new StringItem("", "Generate OTP",StringItem.BUTTON);
         mSpace1 = new Spacer (8,20);
         mSpace2 = new Spacer (8,10);
         generate.setLayout(Item.LAYOUT_CENTER | Item.LAYOUT_NEWLINE_BEFORE);
         generateCommand = new Command("",Command.ITEM,0);
         generate.setDefaultCommand(generateCommand);
         generate.setItemCommandListener(this);

         pinField = new TextField("Enter PIN:", null, 8, TextField.PASSWORD | TextField.NUMERIC);
         pinField.setLayout(Item.LAYOUT_VCENTER);
         mForm.append(mSpace1);
         mForm.append(pinField);
         mForm.append(mSpace2);
         mForm.append(generate);
         mForm.addCommand(exitCommand);
         mForm.setCommandListener(this);

         //waitForm
         Gauge mgauge = new Gauge("Generating password\n\n", false, Gauge.INDEFINITE, Gauge.CONTINUOUS_RUNNING);
         mgauge.setLayout(Item.LAYOUT_CENTER | Item.LAYOUT_VCENTER);
         waitForm.addCommand(cancelCommand);
         waitForm.append(mgauge);
         waitForm.setCommandListener(this);

         //errorForm
         errLine = new StringItem("Error: ", "Incorrect PIN",StringItem.PLAIN);
         errLine.setLayout(Item.LAYOUT_CENTER | Item.LAYOUT_NEWLINE_BEFORE);
         errorForm.append(errLine);
         errorForm.addCommand(backCommand);
         errorForm.setCommandListener(this);

         //hotpForm
         hotpLine = new StringItem("Password: ", "", StringItem.PLAIN);
         hotpLine.setLayout(Item.LAYOUT_CENTER | Item.LAYOUT_NEWLINE_BEFORE);
         hotpForm.append(hotpLine);
         hotpForm.addCommand(exitCommand);
         hotpForm.addCommand(backCommand);
         hotpForm.setCommandListener(this);

         // SVG form that shows the generated password
         try
         {
             msvgForm = null;
             InputStream inStream = getClass().getResourceAsStream(svgname);
             SVGImage svgi = (SVGImage)ScalableImage.createImage(inStream, null);
             Document msvgDoc = svgi.getDocument();
             svgHotp = (SVGElement)msvgDoc.getElementById("hotp_value");
             svgAnime = SVGAnimator.createAnimator(svgi);
             Canvas c = (Canvas)svgAnime.getTargetComponent();
             //c.setFullScreenMode(true);
             svgi.setViewportWidth(c.getWidth());
             svgi.setViewportHeight(c.getHeight());
             msvgForm = c;
             msvgForm.addCommand(exitCommand);
             msvgForm.addCommand(backCommand);
             msvgForm.setCommandListener(this);
         }
         catch (IOException e)
         {msvgForm = null;}
     }//end constructor



     public void startApp()
     {
         mdisplay = Display.getDisplay(this);
         mdisplay.setCurrent(mForm);
     }

     public void pauseApp() {}
     public void destroyApp(boolean unconditional)
     {
         try {cardConnection.close();}
         catch (Exception e){}
     }


     //handling for standard commands
     public void commandAction(Command c, Displayable s)
     {
         if (c == exitCommand)
         {
             try {cardConnection.close();}
             catch (Exception e){}
             
             notifyDestroyed();
         }
         
         else if (c == backCommand)
         {
             pinField.setString("");
             mdisplay.setCurrent(mForm);
         }

         else if (c == cancelCommand)
         {
              try {cardConnection.close();}
              catch (Exception e)
              {
                  Form f = new Form("Exception:");
                  f.append(e.toString());
                  f.addCommand(backCommand);
                  f.setCommandListener(this);
                  mdisplay.setCurrent(f);
              }
              pinField.setString("");
              mdisplay.setCurrent(mForm);
         }
     }//end commandAction


     //handling for commands associated with on-screen Items
     public void commandAction( Command cC, Item cItem )
     {
         if (cC == generateCommand && cItem == generate)
         {
             if (pinField.size() != 0)
             {
                 mdisplay.setCurrent(waitForm);

                 Thread connectionThread = new Thread(this);
                 connectionThread.start();
             }
             else
             {
                 Alert errorAlert = new Alert("Error:", "Please enter PIN", null,AlertType.ERROR);
                 errorAlert.setTimeout(4000);
                 mdisplay.setCurrent(errorAlert);
             }
         }
     }


     // (new thread): creates new APDU connection and exchanges data with the card
     public void run()
     {
         String totp = null;
         if (cardConnection == null)
         {
             try
             {
                 cardConnection = (APDUConnection)Connector.open(appletAID);
             }
             catch (Exception e)
             {
                 cardConnection = null;
                 errLine.setText(e.toString());
                 mdisplay.setCurrent(errorForm);
                 return;
             }
         }

         cardCommunicator.setConnection(cardConnection);

         if (cardCommunicator.verifyPIN(pinField.getString()) == true)
             totp = cardCommunicator.generateTOTP(totpNumberDigits);
         
         if(totp == null)
         {
             errLine.setText("TOTP generation failed");
             mdisplay.setCurrent(errorForm);
             return;
         }

         //show TOTP on the screen
         if (msvgForm != null)//if SVG form was created
         {
             svgHotp.setTrait("#text", totp);// show using scalable vector graphics
             mdisplay.setCurrent (msvgForm);
         }
         else
         {
             hotpLine.setText(totp);// show using ordinary form
             mdisplay.setCurrent (hotpForm);
         }

     }

}
